
class reducer:
    """
    class to be used as a decorator for reducer methods.
    the name of the method is used as the action name, and the
    method is reassigned to a new name with a preceding _

    @reducer
    def MY_ACTION(self, action, state, previous_value):
        # this is a reducer for MY_ACTION, the method is now
        # present on cls._MY_ACTION(), cls.MY_ACTION is now
        # the string "MY_ACTION" (the action type name)
        pass

    if the reducer has args, they are interpreted as the
    state elements to run this reducer for. If there are
    no args it will get run for any state change.
    """
    def __init__(self, *args):
        self.owning_class = None
        self.func = None
        self.states = None
        self.action_name = None
        self.method_name = None

        # if no args to @reducer
        if len(args) == 1 and callable(args[0]):
            self._assign_func(args[0])
        else:
            self.states = args

    def __call__(self, *args, **kwargs):
        if len(args) == 1 and callable(args[0]):
            self._assign_func(args[0])
            return self
        else:
            return self.func(*args, **kwargs)

    def _assign_func(self, func):
        self.func = func
        self.action_name = func.__name__.upper()
        self.method_name = '_' + self.action_name

    # __set_name__ gets called because @reducer is a "descriptor",
    # at the right time in the init process
    def __set_name__(self, owner, name):
        self.owning_class = owner
        setattr(owner, self.method_name, self.func)
        setattr(owner, self.action_name, self.action_name)

        handlers = owner._store_reducers.setdefault(self.action_name, [])

        if self.states is None:
            self.states = self.owning_class.store_attrs

        reducer_id = self.owning_class._next_reducer_id
        self.owning_class._next_reducer_id += 1
        for state in self.states:
            handlers.append((reducer_id, state, self))



import asyncio

from flopsy.store import SyncedStore
from flopsy.reducer import reducer
from flopsy.saga import saga


class StateObject(SyncedStore):
    store_attrs = ["xpos", "ypos"]
    store_type = "StateObject"
    store_id_attr = "id"

    _next_id = 1

    def __init__(self):
        self.id = StateObject._next_id
        StateObject._next_id += 1

        self.xpos = None
        self.ypos = None

        super().__init__()

    @reducer
    def clear_pos(self, action, state, oldval):
        return 0

    @reducer('xpos')
    def incr_xpos(self, action, state, oldval):
        return oldval + 1

    @reducer('ypos')
    def incr_ypos(self, action, state, oldval):
        return oldval + 1

    @saga('xpos')
    async def drip_incr_x(self, action, state_diff):
        for _ in range(5):
            await asyncio.sleep(1)
            yield self.action(StateObject.INCR_YPOS)


ss = StateObject()
inspector = ss.show_inspector()

await ss.action(StateObject.CLEAR_POS).dispatch()

for i in range(20):
    await asyncio.sleep(1)
    await ss.action(StateObject.INCR_YPOS).dispatch()





class Action:
    """
    Actions change the state of the target when processed by a reducer
    """
    def __init__(self, target, type_name, payload):
        self.target = target
        self.type_name = type_name
        self.payload = payload

    def dispatch(self):
        return self.target.dispatch(self)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import asyncio
import json
import copy
from datetime import datetime
from threading import Thread
from imgui.integrations.glfw import GlfwRenderer
import OpenGL.GL as gl
import glfw
import imgui
from flopsy.action import Action
from flopsy.store import Store


class Inspector(Thread):
    def __init__(self, *, width=800, height=600):
        super().__init__()
        self.window_name = "Flopsy inspector"
        self.window_width = width
        self.window_height = height

        self.timeline_item_selected = None
        self.store_item_selected = None
        self.store_item_selected_value = None

        self.sagas = {}
        self.timeline = []

        self.event_loop = asyncio.get_event_loop()

        for store_type in Store.all_store_types():
            self.sagas[store_type.__name__] = store_type.install_saga(
                self.update_timeline,
            )

        self.initial_store = Store.store()
        self.initial_store_ts = datetime.now()
        self.current_store = copy.copy(self.initial_store)

    async def update_timeline(self, store, action, state_diff):
        self.timeline.append([
            datetime.now(),
            action,
            state_diff
        ])
        if self.timeline_item_selected is None:
            store_items = self.current_store.setdefault(store.store_type, {})
            store_item_content = store_items.setdefault(store.id, {})

            for state_key, values in state_diff.items():
                store_item_content[state_key] = values[1]
        yield None

    def store_dispatch_change(self, store_type, store_id, attr, value):
        store = Store.find_store(store_type, store_id)
        action = Action(
            target=store,
            type_name=f"SET_{attr.upper()}",
            payload=dict(value=value)
        )
        asyncio.run_coroutine_threadsafe(action.dispatch(), self.event_loop)

    def update_timeline_selection(self, new_position):
        last_position = self.timeline_item_selected
        self.timeline_item_selected = new_position

        if last_position is None:
            self.current_store = copy.copy(self.initial_store)
            last_position = 0

        if new_position >= last_position:
            incr = 1
            new_position += 1
        else:
            incr = -1

        for p in range(last_position, new_position, incr):
            store = self.timeline[p][1].target
            store_items = self.current_store.setdefault(store.store_type, {})
            store_item_content = store_items.setdefault(store.id, {})
            for state_key, values in self.timeline[p][2].items():
                new_value = values[1] if incr > 0 else values[0]
                store_item_content[state_key] = new_value

    def create_glfw_window(self):
        if not glfw.init():
            print("Could not initialize OpenGL context")
            return None

        # OS X supports only forward-compatible core profiles from 3.2
        glfw.window_hint(glfw.CONTEXT_VERSION_MAJOR, 3)
        glfw.window_hint(glfw.CONTEXT_VERSION_MINOR, 3)
        glfw.window_hint(glfw.OPENGL_PROFILE, glfw.OPENGL_CORE_PROFILE)

        glfw.window_hint(glfw.OPENGL_FORWARD_COMPAT, gl.GL_TRUE)

        # Create a windowed mode window and its OpenGL context
        window = glfw.create_window(
            int(self.window_width),
            int(self.window_height),
            self.window_name,
            None,
            None
        )
        glfw.make_context_current(window)

        if not window:
            glfw.terminate()
            print("Could not initialize Window")
            return None

        return window

    def imgui_paint(self):
        keep_going = True

        ########################################
        # global menu bar
        if imgui.begin_main_menu_bar():
            if imgui.begin_menu("File"):

                clicked, enabled = imgui.menu_item("Clear timeline")
                if clicked:
                    self.timeline = []
                    self.current_store = Store.store()
                    self.timeline_item_selected = None

                clicked, enabled = imgui.menu_item("Follow timeline")
                if clicked:
                    self.current_store = Store.store()
                    self.timeline_item_selected = None

                # Quit
                clicked, rest = imgui.menu_item("Quit", "Ctrl+Q")
                if clicked:
                    keep_going = False
                imgui.end_menu()

            imgui.end_main_menu_bar()

        # one window that fills the workspace
        imgui.set_next_window_size(self.window_width, self.window_height-21)
        imgui.set_next_window_position(0, 21)
        imgui.get_style().window_rounding = 0
        imgui.style_colors_light()

        imgui.begin(
            "Flopsy Inspector",
            closable=False,
            flags=imgui.WINDOW_NO_MOVE| imgui.WINDOW_ALWAYS_AUTO_RESIZE,
        )

        ########################################
        # timeline
        halfwidth = imgui.get_window_width() * 0.5
        fullheight = imgui.get_window_height()
        imgui.set_next_window_size(halfwidth, fullheight)
        imgui.set_next_window_position(0, 21)
        imgui.begin(
            "Timeline",
            closable=False,
            flags=imgui.WINDOW_NO_COLLAPSE|imgui.WINDOW_NO_MOVE|imgui.WINDOW_NO_RESIZE,
        )
        items = [
            (
                f"{ts.strftime('%H:%M:%S.%f')[:-3]} {action.type_name}",
                action.target,
                action.payload,
                state_diff
            )
            for ts, action, state_diff in self.timeline
        ]
        imgui.begin_child("##timeline-list")

        for counter, item in enumerate(items):
            i_label, i_target, i_payload, i_state_diff = item
            flags = imgui.TREE_NODE_OPEN_ON_DOUBLE_CLICK
            if self.timeline_item_selected == counter:
                flags |= imgui.TREE_NODE_SELECTED
            opened = imgui.tree_node(f"{i_label}##{counter}", flags)
            if imgui.is_item_clicked():
                self.update_timeline_selection(counter)
            if opened:
                imgui.text(f"Store: {type(i_target).__name__}:{i_target.id}")
                imgui.text(f"Payload: {json.dumps(i_payload, indent=4)}")

                if imgui.tree_node(f"State diff##{counter}"):
                    for store_attr, store_diff in i_state_diff.items():
                        imgui.text(f"{store_attr}:")
                        imgui.same_line()
                        imgui.text(f"{store_diff}")
                    imgui.tree_pop()

                imgui.tree_pop()
        imgui.end_child()

        imgui.end()

        ########################################
        # store
        imgui.set_next_window_size(halfwidth, fullheight)
        imgui.set_next_window_position(halfwidth, 21)
        imgui.begin(
            "Store",
            closable=False,
            flags=imgui.WINDOW_NO_COLLAPSE|imgui.WINDOW_NO_MOVE|imgui.WINDOW_NO_RESIZE,
        )

        if self.timeline_item_selected is not None:
            ts = self.timeline[self.timeline_item_selected][0]
        elif self.timeline:
            ts = self.timeline[-1][0]
        else:
            ts = self.initial_store_ts
        imgui.text(f"Last update: {ts.strftime('%H:%M:%S.%f')[:-3]}")

        imgui.begin_child("##store-list")
        for store_type, store_objmap in self.current_store.items():
            if imgui.tree_node(store_type):
                for obj_id, obj_store in store_objmap.items():
                    if imgui.tree_node(f"{obj_id}"):
                        for store_attr, store_value in obj_store.items():
                            imgui.text(f"{store_attr}:")
                            imgui.same_line()
                            item_id = f"{store_type}:{obj_id}:{store_attr}"
                            if self.store_item_selected == item_id:
                                imgui.push_item_width(
                                    0.5 * imgui.get_window_width()
                                )
                                changed, self.store_item_selected_value = imgui.input_text(
                                    "##store_edit_input",
                                    self.store_item_selected_value
                                )
                                imgui.pop_item_width()
                                imgui.same_line()
                                if imgui.button("Change"):
                                    self.store_dispatch_change(
                                        store_type, obj_id, store_attr,
                                        self.store_item_selected_value
                                    )
                                    self.store_item_selected = None
                            else:
                                selected, _ = imgui.selectable(f"{store_value}")
                                if selected:
                                    self.store_item_selected = item_id
                                    self.store_item_selected_value = str(store_value)

                        imgui.tree_pop()
                imgui.tree_pop()
        imgui.end_child()

        imgui.end()

        imgui.end()
        return keep_going

    def run(self):
        # set up the window and renderer context
        imgui.create_context()
        window = self.create_glfw_window()
        impl = GlfwRenderer(window)

        keep_going = True

        while keep_going and not glfw.window_should_close(window):
            # top of loop stuff
            glfw.poll_events()
            impl.process_inputs()
            imgui.new_frame()

            width, height = glfw.get_window_size(window)
            self.window_width = width
            self.window_height = height

            # hard work
            keep_going = self.imgui_paint()

            # bottom of loop stuff
            gl.glClearColor(1.0, 1.0, 1.0, 1)
            gl.glClear(gl.GL_COLOR_BUFFER_BIT)
            imgui.render()
            impl.render(imgui.get_draw_data())
            glfw.swap_buffers(window)

        impl.shutdown()
        glfw.terminate()




import asyncio
from .action import Action


class Store:
    """
    Base class for stores of managed state

    Class attributes configure behavior:

    store_attrs: array of store attribute names. Each one
    will get an automatic class attribute and setter action.
    if the attr is "my_attr" in class MyStore, there will be
    a MyStore.MY_ATTR which is the string "my_attr", and
    an action type MyStore.SET_MY_ATTR with an automatic
    reducer.

    store_type: the type name for the store to be used
    when aggregating the application's stores. Defaults
    to the class __name__

    store_id_attr: the attribute to use as a type-unique
    id for the store, to be used when aggregating the
    application's stores. Defaults to "id"
    """
    _store_registry = {}
    _store_reducers = {}
    _store_sagas = []
    _store_tasks = []

    _next_saga_id = 1
    _next_reducer_id = 1

    @classmethod
    def _setter_helper(cls, attr):
        def inner(self, newval):
            oldval = getattr(self, attr)
            setattr(self, attr, newval)
            return (attr, oldval)
        return inner

    def __init_subclass__(cls):
        if not hasattr(cls, 'store_attrs'):
            return

        Store._store_registry[cls.__name__] = {}

        for attr in cls.store_attrs:
            setter_action = f"SET_{attr.upper()}"
            setter_methodname = f"_{setter_action}"
            setter_dispatch = cls._setter_helper(attr)

            getter_statename = f"{attr.upper()}"

            # define the name as a symbol
            setattr(cls, setter_action, setter_action)
            setattr(cls, getter_statename, attr)

            # define the setter as a method
            setattr(cls, setter_methodname, setter_dispatch)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        id_attr = getattr(self, "store_id_attr", "id")
        Store._store_registry[type(self).__name__][getattr(self, id_attr)] = self

    def action(self, action_label, **kwargs):
        return Action(self, action_label, kwargs)

    def state(self, label=None):
        if not label:
            return {
                attr: getattr(self, attr)
                for attr in self.store_attrs
            }
        return getattr(self, label)

    async def dispatch(self, action):
        handlers = self._store_reducers.get(action.type_name, [])
        state_diff = {}
        for callback_id, state_name, cb in handlers:
            old_value = getattr(self, state_name)
            new_value = cb(self, action, state_name, old_value)
            setattr(self, state_name, new_value)
            if old_value != new_value:
                state_diff[state_name] = (old_value, new_value)

        if not handlers:
            magic_handler_name = f"_{action.type_name}"
            if hasattr(type(self), magic_handler_name):
                magic_handler = getattr(self, magic_handler_name)
                new_value = action.payload.get("value", None)
                state_name, old_value = magic_handler(new_value)
                if old_value != new_value:
                    state_diff[state_name] = (old_value, new_value)

        changed_state_set = set(state_diff.keys())
        if not changed_state_set:
            return

        for callback_id, callback, state_filter in self._store_sagas:
            if state_filter and not (set(state_filter) & changed_state_set):
                continue
            self._launch_task(
                self._run_saga(
                    callback(self, action, state_diff)
                )
            )

    def _launch_task(self, task):
        # clean up completed tasks
        Store._store_tasks = [
            t for t in Store._store_tasks
            if t and not t.done()
        ]

        # save the new one
        Store._store_tasks.append(asyncio.create_task(task))

    async def _run_saga(self, saga):
        async for action in saga:
            if action and isinstance(action, Action):
                await self.dispatch(action)

    # install reducer for states
    @classmethod
    def install_reducer(cls, action_name, states):

        reducer_id = Store._next_reducer_id
        Store._next_reducer_id += 1

        handlers = cls._store_reducers.setdefault(action_name, [])
        reducer = getattr(cls, f"_{action_name}")

        for state in states:
            handlers.append((reducer_id, state, reducer))
        return reducer_id

    @classmethod
    def uninstall_reducer(cls, action_name, reducer_id):
        handlers = cls._store_reducers.setdefault(action_name, [])
        handlers[:] = [
            h for h in handlers
            if h[0] != reducer_id
        ]

    @classmethod
    def install_saga(cls, saga, states=None):
        saga_id = Store._next_saga_id
        Store._next_saga_id += 1

        if not states:
            states = []

        cls._store_sagas.append((saga_id, saga, states))
        return saga_id

    @classmethod
    def uninstall_saga(cls, saga_id):
        cls._store_sagas = [
            h for h in cls._store_sagas
            if h[0] != saga_id
        ]

    @staticmethod
    def all_store_types():
        stores = []
        for typename, type_objects in Store._store_registry.items():
            if len(type_objects):
                stores.append(type(next(iter(type_objects.values()))))
        return stores

    @staticmethod
    def all_stores():
        stores = []
        for typename, type_objects in Store._store_registry.items():
            stores = stores + list(type_objects.values())
        return stores

    @staticmethod
    def store():
        complete_state = {}
        for typename, type_objects in Store._store_registry.items():
            typereg = complete_state.setdefault(typename, {})
            for obj_id, obj in type_objects.items():
                typereg[obj_id] = obj.state()
        return complete_state

    @staticmethod
    def find_store(store_type, store_id):
        return Store._store_registry.get(store_type, {}).get(store_id, None)

    @staticmethod
    def show_inspector():
        from .inspector.inspector import Inspector
        inspector = Inspector()
        inspector.start()
        return inspector

class SyncedStore (Store):
    """
    A type of store with convenience actions for syncing
    with a remote source of truth.

    In addition to the normal setters, attr "my_attr" will
    also have a secondary setter action called
    MyStore.SYNC_MY_ATTR. This behaves like the setter in
    every way, except for the action type.

    This can be useful if you have an after_dispatch handler
    that pushes state to the remote source of truth. If the
    action is a SYNC_* one, you know you are receiving
    a state update so you don't need to push it.
    """
    def __init_subclass__(cls):
        super().__init_subclass__()

        if not hasattr(cls, 'store_attrs'):
            return

        for attr in cls.store_attrs:
            setter_action = f"SYNC_{attr.upper()}"
            setter_methodname = f"_{setter_action}"
            setter_dispatch = cls._setter_helper(attr)

            # define the name as a symbol
            setattr(cls, setter_action, setter_action)

            # define the setter as a method
            setattr(cls, setter_methodname, setter_dispatch)





# build with 'python ./setup.py install'
from distutils.core import setup

VERSION = "0.0.1"

setup(
    name = 'flopsy',
    version = VERSION,
    license = 'MIT',
    description = 'Redux-inspired state management',
    author = 'Bill Gribble',
    author_email = 'grib@billgribble.com',
    url = 'https://github.com/bgribble/flopsy',
    download_url = 'https://github.com/bgribble/flopsy/archive/refs/tags/0.0.1.tar.gz',
    keywords = ['state-management', 'redux', 'saga'],
    install_requires = [
        'pyopengl', 'glfw', 'imgui[glfw]',
    ],
    classifiers = [
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License',
    ],
)




class saga:
    """
    class to be used as a decorator for saga methods.
    the name of the method is used as the saga name, and the
    method is reassigned to a new name with a preceding _

    @saga
    def my_saga(self, action, state_diff):
        # this is a saga that runs after all dispatchers
        # for "action". It should be an async generator --
        # that is, an async function that uses yield to
        # return a sequence of values.
        pass

    if the reducer has args, they are interpreted as the
    state elements. The saga will only be run after an action
    that affects the named state items. If there are
    no args it will get run after any action.
    """
    def __init__(self, *args):
        self.owning_class = None
        self.func = None
        self.states = None
        self.action_name = None
        self.method_name = None

        # if no args to @reducer
        if len(args) == 1 and callable(args[0]):
            self._assign_func(args[0])
        else:
            self.states = args

    def __call__(self, *args, **kwargs):
        if len(args) == 1 and callable(args[0]):
            self._assign_func(args[0])
            return self
        else:
            return self.func(*args, **kwargs)

    def _assign_func(self, func):
        self.func = func
        self.action_name = func.__name__.upper()
        self.method_name = '_' + self.action_name

    # __set_name__ gets called because @reducer is a "descriptor",
    # at the right time in the init process
    def __set_name__(self, owner, name):
        self.owning_class = owner
        setattr(owner, self.method_name, self.func)
        setattr(owner, self.action_name, self.action_name)

        if self.states is None:
            self.states = self.owning_class.store_attrs
        saga_id = self.owning_class._next_saga_id
        self.owning_class._next_saga_id += 1
        self.owning_class._store_sagas.append((saga_id, self, self.states))

